minispade.register('ember-timetree/views/timetree_view', "(function() {// Copyright (C) 2013 CrowdStrike, Inc. and contributors\n// This file is subject to the terms and conditions of the BSD License.\n// See the file LICENSE in the main directory for details\n\nfunction lastEndTime(node, count) {\n  if (!count) { count = 0; }\n  if (count > 100) { throw new Error(\"Infinite loop in children when finding lastEndTime\"); }\n\n  var children = node.children || node._children,\n      lastEnd = node.end;\n  if (children) {\n    children.forEach(function(child) {\n      var childLastEnd = lastEndTime(child, count+1);\n      if (childLastEnd > lastEnd) {\n        lastEnd = childLastEnd;\n      }\n    });\n  }\n  return lastEnd;\n}\n\nEmber.Timetree.TimetreeView = Ember.View.extend({\n  tagName: 'svg',\n\n  width: 750,\n  rowHeight: 15,\n  rowSpacing: 10,\n  labelsWidth: 200,\n  axisHeight: 20,\n  axisPosition: 'bottom',\n  indentSize: 20,\n  labelAlign: 'left',\n  contentMargin: null, // { top: 0, left: 0, bottom: 0, right: 0 },\n\n  collapsable: true,\n  scrubbable: true,\n  selectable: true,\n  brushable: false,\n  resizeOnCollapse: false,\n\n  showLabels: true,\n  showLinks: true,\n\n  attributeBindings: ['_width:width', 'height'],\n\n  classNames: ['timetree-view'],\n\n  /**\n    [\n      parent:   null, // index of parent in array\n      label:    \"Name\",\n      start:    12345,\n      end:      67890,\n\n      // id for determining uniqueness, optional, defaults to index\n      id: null,\n\n      // Class name applied to labels and bars, optional\n      className: null,\n\n      // Bar sections to apply different classes to, optional\n      sections: [{ start: 12345, end: 23456, className: 'active' },\n                 { start: 23456, end: 67890, className: 'inactive' }],\n\n      // Content for selection, optional. If empty, the outer object is selected.\n      content:  {Obj}\n    ]\n  */\n  content: null,\n\n  selection: null,\n\n  brushRange: null,\n  range: null,\n\n  _translateRegex: /translate\\(([\\d\\.]+),?\\s*([\\d\\.]+)\\)/,\n\n  minimumWidth: Ember.computed(function() {\n    return this.get('labelsWidth') + 100;\n  }).property('labelsWidth'),\n\n  maximumWidth: Ember.computed(function() {\n    if (this.get('element')) {\n      return this.$().parent().width();\n    } else {\n      return 0;\n    }\n  }).property('element').volatile(),\n\n  _width: Ember.computed(function() {\n    var width = this.get('width');\n    if (width === 'auto') { width = this.get('maximumWidth'); }\n    return Math.max(this.get('minimumWidth'), width);\n  }).property('width', 'maximumWidth', 'minimumWidth'),\n\n  barsHeight: Ember.computed(function() {\n    return (this.get('rowHeight') + this.get('rowSpacing')) * (this.get(\"resizeOnCollapse\") ? this.get(\"visibleNodeCount\") : this.get('content.length'));\n  }).property('rowHeight', 'rowSpacing', 'content.length', 'visibleNodeCount', 'resizeOnCollapse'),\n\n  contentHeight: Ember.computed(function() {\n    return this.get('barsHeight') + (this.get('contentMargin.top') || 0) + (this.get('contentMargin.bottom') || 0);\n  }).property('barsHeight'),\n\n  height: Ember.computed(function() {\n    return this.get('axisHeight') + this.get('contentHeight');\n  }).property('axisHeight', 'contentHeight'),\n\n  _range: Ember.computed(function() {\n    var range = this.get('range');\n    if (range && range.length === 2) { return range; }\n\n    // Some of this is a bit redundant with renderNodes\n    var rootNode = this.get('rootNode');\n    if (rootNode.children.length === 0) { return []; }\n\n    var tree = this.get('tree'),\n        nodes = tree.nodes(rootNode).slice(1), // Skip root node\n        min = d3.min(nodes, function(n) { return n.start; }),\n        max = d3.max(nodes, function(n) { return n.lastEnd; });\n\n    return [min, max];\n  }).property('range', 'rootNode'),\n\n  _labelAt: function(i) { return d3.select(this.get('svg').selectAll('.labels .label')[0][i]); },\n\n  contentWidth: Ember.computed(function() {\n    return this.get('_width') - this.get('labelsWidth');\n  }).property('_width', 'labelsWidth'),\n\n  barsWidth: Ember.computed(function() {\n    return this.get('contentWidth') - (this.get('contentMargin.left') || 0) - (this.get('contentMargin.right') || 0);\n  }).property('contentWidth', 'contentMargin.left', 'contentMargin.right'),\n\n  timeFormat: Ember.computed(function() {\n    return d3.time.format.utc(\"%H:%M:%S.%L\");\n  }).property(),\n\n  timeTickFormat: Ember.computed(function() {\n    var timeFormat = this.get('timeFormat');\n    return function(d){ return timeFormat(new Date(d)); };\n  }).property('timeFormat'),\n\n  xScale: Ember.computed(function() {\n    return d3.scale.linear().range([0, this.get('contentWidth')]).clamp(true);\n  }).property(),\n\n  yScale: Ember.computed(function() {\n    var rowHeight = this.get('rowHeight'),\n        rowSpacing = this.get('rowSpacing'),\n        paddingFactor = rowSpacing/(rowHeight+rowSpacing);\n    return d3.scale.ordinal().rangeRoundBands([0, this.get('barsHeight')], paddingFactor, paddingFactor/2);\n  }).property(),\n\n  xAxis: Ember.computed(function() {\n    return d3.svg.axis()\n              .scale(this.get('xScale'))\n              .orient(\"bottom\")\n              .tickFormat(this.get('timeTickFormat'))\n              .ticks(5);\n  }).property(),\n\n  brush: Ember.computed(function() {\n    return d3.svg.brush()\n              .x(this.get('xScale'))\n              .on('brush', Ember.$.proxy(this, 'doBrush'));\n  }).property(),\n\n  tree: Ember.computed(function() {\n    return d3.layout.tree();\n  }).property(),\n\n  svg: Ember.computed(function() {\n    var element = this.get('element');\n    return element ? d3.select(element) : null;\n  }).property('element'),\n\n  rootNode: Ember.computed(function() {\n    var nodes = this.get('content'),\n        rootNode = { label: 'root', children: [] };\n\n    if (nodes) {\n      nodes = nodes.map(function(node) {\n        return Ember.$.extend({}, node);\n      });\n\n      nodes.forEach(function(node) {\n        var parentNode = node.parent != null ? nodes[node.parent] : rootNode;\n        if (!parentNode.children) { parentNode.children = []; }\n        parentNode.children.push(node);\n      });\n\n      nodes.forEach(function(node, idx) {\n        if (!node.id) { node.id = idx; }\n        if (!node.className) { node.className = ''; }\n        node.lastEnd = lastEndTime(node);\n      });\n    }\n\n    return rootNode;\n  }).property('content'),\n\n  visibleNodeCount: Ember.computed(function(){\n    var total = 1,\n        computeTotal = function(node){\n          for( var i = 0; i < node.length; i++){\n            var n = node[i];\n            if( n.children ){\n              total += n.children.length;\n              computeTotal( n.children);\n            }\n          }\n        };\n\n    computeTotal([this.get(\"rootNode\")]);\n    return total;\n  }).property(\"rootNode\"),\n\n  adjustXScaleRange: Ember.observer(function() {\n    this.get('xScale').range([0, this.get('barsWidth')]);\n  }, 'barsWidth'),\n\n  adjustYScaleRange: Ember.observer(function() {\n    // TODO: This is a copy of `yScale`, clean it up\n    var rowHeight = this.get('rowHeight'),\n        rowSpacing = this.get('rowSpacing'),\n        paddingFactor = rowSpacing/(rowHeight+rowSpacing);\n    this.get('yScale').rangeRoundBands([0, this.get('barsHeight')], paddingFactor, paddingFactor/2);\n  }, 'barsHeight', 'rowHeight', 'rowSpacing'),\n\n  adjustScrubberHeight: Ember.observer(function() {\n    if (this.get('scrubbable')) {\n      var height = this.get('contentHeight'),\n          scrubber = this.get('svg').select('.scrubber');\n      scrubber.select('line').attr('y2', height);\n      scrubber.select('text').attr('y', height);\n    }\n  }, 'contentHeight'),\n\n  updateXAxisScale: Ember.observer(function() {\n    this.get('xAxis').scale(this.get('xScale'));\n  }, 'xScale'),\n\n  updateRows: function(rowItems) {\n    var yScale = this.get('yScale').copy(),\n        width = this.get('_width'),\n        height = this.get('barsHeight');\n\n    yScale.rangeRoundBands([0, height], 0, 0);\n\n    rowItems\n        .attr('x', 0)\n        .attr('y', function(n,i) { return yScale(i); })\n        .attr('width', width)\n        .attr('height', yScale.rangeBand());\n  },\n\n  drawAxis: function() {\n    this.get('svg').select(\".x.axis\").call(this.get('xAxis'));\n  },\n\n// This function is passed to d3, it's not called as a member of the View\n  durationFormatter: function(n) {\n    return (n.end - n.start)/1000 + 's';\n  },\n\n  renderNodes: function() {\n    var rootNode = this.get('rootNode');\n    if (rootNode.children.length === 0) { return; }\n\n    var tree = this.get('tree'),\n        nodes = tree.nodes(rootNode).slice(1); // Skip root node\n\n    var width = this.get('_width'),\n        labelsWidth = this.get('labelsWidth'),\n        leftPadding = this.get('contentMargin.left') || 0,\n        svg = this.get('svg'),\n        rows = svg.select('.rows'),\n        labels = svg.select('.labels'),\n        timeTickFormat = this.get('timeTickFormat'),\n        collapsable = this.get('collapsable'),\n        scrubbable = this.get('scrubbable'),\n        brushable = this.get('brushable'),\n        showLabels = this.get('showLabels'),\n        showLinks = this.get('showLinks'),\n        durationFormatter = this.get('durationFormatter'),\n        applyLabel,\n        range = this.get('_range');\n\n    var xScale = this.get('xScale'),\n        yScale = this.get('yScale'),\n        xAxis = this.get('xAxis'),\n        content = svg.select('.content');\n\n    xScale.domain(range);\n    yScale.domain(d3.range(this.get(\"resizeOnCollapse\") ? this.get(\"visibleNodeCount\") : this.get('content.length')));\n\n    var rowItems = rows.selectAll('.row')\n                        .data(nodes, function(n) { return n.id; });\n\n    rowItems.enter().append('rect')\n        .attr('class', function(n) { return 'row '+n.className; });\n\n    rowItems.exit().remove();\n\n    this.updateRows(rowItems);\n\n    if (labelsWidth > 0) {\n      var indentSize = this.get('indentSize'),\n          labelAlign = this.get('labelAlign'),\n          showCircles;\n\n      showLinks = showLinks && indentSize > 0;\n      showCircles = collapsable || showLinks;\n\n      if (showLinks) {\n        var links = tree.links(nodes),\n            linkItems = labels.selectAll(\".link\")\n                            .data(links);\n\n        linkItems.enter().insert(\"path\", \":first-child\")\n            .attr(\"class\", \"link\");\n\n        linkItems.exit().remove();\n\n        linkItems.attr(\"d\", function(d) {\n              return \"M\" + (d.source.depth * indentSize) + \",\" + (yScale(nodes.indexOf(d.source)) + yScale.rangeBand() / 2) +\n                     \"V\" + (yScale(nodes.indexOf(d.target)) + yScale.rangeBand() / 2) + \"H\" + (d.target.depth * indentSize);\n            });\n      }\n\n      var labelItems = labels.selectAll('.label')\n                          .data(nodes, function(n) { return n.id; });\n\n      var labelItemsEnter = labelItems.enter().append('g')\n            .attr('class', function(n) { return 'label '+n.className; });\n\n      if (showCircles) {\n        var circles = labelItemsEnter.append('circle');\n\n        if (collapsable) {\n          var self = this;\n          circles.attr('class', 'collapsable')\n            .on('click', function(n) {\n              self.toggleNode(n);\n              self.renderNodes();\n            });\n          // update after initial draw if nodes changed\n          labelItems.selectAll(\"circle\").data(nodes, function(n) { return n.id; });\n        }\n      }\n\n      labelItemsEnter.append('text');\n\n      labelItems.exit().remove();\n\n      labelItems\n        .attr('transform', function(n,i) { return 'translate('+ (n.depth * indentSize) + ',' + (yScale(i) + yScale.rangeBand() / 2) + ')'; });\n\n      labelItems\n        .classed('has-children', function(n) { return n.children || n._children; });\n\n      labelItems\n        .classed('closed', function(n) { return n._children; });\n\n      labelItems.selectAll('circle')\n        .attr('cx', 0)\n        .attr('cy', 0)\n        .attr('r', 6);\n\n      var labelText = labelItems.selectAll('text');\n\n      labelText\n        .attr('dx', showLinks ? 10 : 0) // padding-left\n        .attr('dy', \".35em\") // vertical-align: middle\n        .text(function(n) { return n.label; });\n\n      if (labelAlign === 'right') {\n        labelText\n          .attr(\"text-anchor\", \"end\")\n          .attr('x', function(n) { return labelsWidth + xScale(n.start) - 10; });\n      }\n    }\n\n    this.drawAxis();\n\n    var bars = content.selectAll('.bars').selectAll('.bar')\n                      .data(nodes, function(n) { return n.id; });\n\n    var barsEnter = bars.enter().append('g')\n                            .attr('class', function(n) { return 'bar ' + n.className + ' ' + (n.sections ? 'sectional' : ''); });\n\n    var barsEnterDurationGroup = barsEnter.append('g')\n                                            .attr('class', 'whole duration');\n\n    barsEnterDurationGroup.append('rect');\n    barsEnterDurationGroup.append('text');\n\n    barsEnter.call(function(sel) {\n      sel.each(function(n, i) {\n        var sectionGroup;\n        if (n.sections) {\n          sectionGroup = d3.select(this).selectAll('.section.duration').data(n.sections).enter()\n            .append('g')\n              .attr('class', function(s) { return 'section duration ' + (s.className || ''); });\n\n          sectionGroup.append('rect');\n          sectionGroup.append('text');\n        }\n      });\n    });\n\n    applyLabel = function(selection) {\n      return selection\n        .attr('y', function() { return yScale.rangeBand() / 2; })\n        .attr('dx', 3) // padding-left\n        .attr('dy', \".35em\") // vertical-align: middle\n        .text(durationFormatter);\n    };\n\n    bars.exit().remove();\n\n    bars\n        .attr(\"transform\", function(n,i){ return \"translate(\" + (xScale(n.start) + leftPadding) + \",\" + yScale(i) + \")\"; })\n        .classed(\"collapsed\", function(n) {\n          return n._children;\n        });\n\n\n    bars.selectAll('.duration rect')\n        .attr('width', function(n) { return xScale(n.end) - xScale(n.start); })\n        .attr('height', yScale.rangeBand());\n\n    bars.call(function(sel) {\n      sel.each(function(n, i) {\n        var sectionData;\n        if (n.sections) {\n          sectionData = d3.select(this).selectAll('.section').data(n.sections);\n          sectionData.select('rect')\n              .attr('x', function(s) { return xScale(s.start) - xScale(n.start); })\n              .attr('width', function(s) { return xScale(s.end) - xScale(s.start); })\n              .attr('height', yScale.rangeBand());\n\n          if (showLabels) {\n            sectionData.select('text')\n              .attr('x', function(s) { return xScale(s.start) - xScale(n.start); })\n              .call(applyLabel);\n          }\n        }\n      });\n    });\n\n    if (showLabels) {\n      bars.selectAll('.whole.duration text')\n        .attr('x', 0)\n        .call(applyLabel);\n    }\n\n    if (scrubbable) {\n      this.doScrub(this._currentScrubberX());\n    }\n\n    if (brushable) {\n      var brush = this.get('brush'),\n          contentHeight = this.get('contentHeight');\n\n      content.selectAll('.brush')\n          .call(brush)\n        .selectAll(\"rect\")\n          .attr(\"height\", contentHeight);\n    }\n\n    if (this.didRenderNodes) {\n      this.didRenderNodes(nodes);\n    }\n  },\n\n  nodesNeedRerender: Ember.observer(function() {\n    Ember.run.once(this, 'renderNodes');\n  }, 'rootNode', 'height', '_width', '_range.[]'),\n\n  doHighlight: function(y) {\n    var self = this, svg;\n\n    if (!this.get('selectable')) { return; }\n\n    svg = this.get('svg');\n    svg.select('.rows').selectAll('.row')\n      .classed('hover', function(d, i) {\n        var top = Number(this.getAttribute('y')),\n            bottom = top + Number(this.getAttribute('height')),\n            isSelected = top <= y && y < bottom;\n\n        self._labelAt(i).classed('hover', isSelected);\n\n        return isSelected;\n      });\n  },\n\n  doScrub: function(x) {\n    if (!this.get('scrubbable')) { return; }\n\n    var svg = this.get('svg'),\n        scrubber = svg.select('.content').select('.scrubber'),\n        xScale = this.get('xScale'),\n        contentWidth = this.get('contentWidth'),\n        leftPadding = this.get('contentMargin.left') || 0,\n        self = this;\n\n    if (x === undefined) { x = 0; }\n\n    scrubber\n        .attr('transform', 'translate('+x+', 0)');\n\n    var text = scrubber.select('text')\n                       .attr('dy', -1);\n\n    text.text(x > 0 ? this.get(\"timeTickFormat\")(xScale.invert(x-leftPadding)) : \"\");\n\n    var textWidth = text.node().getComputedTextLength ? text.node().getComputedTextLength() : 0;\n    if (x + textWidth * 1.5 > contentWidth) {\n      scrubber.classed('switched', true);\n      text.attr(\"text-anchor\", 'end')\n          .attr('x', -4);\n    } else {\n      scrubber.classed('switched', false);\n      text.attr(\"text-anchor\", 'start')\n          .attr('x', 4);\n    }\n\n    var bars = svg.select('.content').selectAll('.bar');\n    bars.each(function() {\n      // FIXME: This should be data based\n      var transform = this.getAttribute('transform'),\n          match = self._translateRegex.exec(transform),\n          relStart = Number(match[1]);\n\n      d3.select(this).selectAll('.duration').classed('hover', function() {\n        var rect = d3.select(this).select('rect'),\n            start = relStart + Number(rect.attr('x') || 0),\n            end = start + Number(rect.attr('width'));\n        return start <= x && x <= end;\n      });\n    });\n  },\n\n  _currentScrubberX: function() {\n    var scrubber = this.get('svg').select('.scrubber'),\n        transform = !scrubber.empty() && scrubber.attr('transform'),\n        match = this._translateRegex.exec(transform) || [],\n        translateX = match[1];\n\n    return translateX ? Number(translateX) : undefined;\n  },\n\n  doBrush: function() {\n    if (!this.get('brushable')) { return; }\n\n    var brush = this.get('brush');\n\n    this.set('brushRange', brush.empty() ? null : brush.extent());\n  },\n\n  toggleNode: function(node) {\n    if (node.children) {\n      node._children = node.children;\n      delete node.children;\n    } else {\n      node.children = node._children;\n      delete node._children;\n    }\n    this.notifyPropertyChange(\"visibleNodeCount\");\n  },\n\n  didInsertElement: function() {\n    var labelsWidth = this.get('labelsWidth'),\n        contentWidth = this.get('contentWidth'),\n        contentHeight = this.get('contentHeight'),\n        scrubbable = this.get('scrubbable'),\n        selectable = this.get('selectable'),\n        brushable = this.get('brushable'),\n        axisHeight = this.get('axisHeight'),\n        axisPosition = this.get('axisPosition'),\n        contentTop = axisPosition === 'top' ? axisHeight : 0,\n        axisTop = axisPosition === 'top' ? axisHeight : contentHeight;\n\n    this.adjustScrubberHeight();\n\n    var svg = this.get('svg'),\n        rows, labels, content, scrubber;\n\n    var self = this;\n\n    if (axisHeight > 0) {\n      svg.append(\"g\")\n        .attr(\"class\", \"x axis\")\n        .attr(\"transform\", \"translate(\" + labelsWidth + \",\" + axisTop + \")\");\n    }\n\n    rows = svg.insert(\"g\", \":first-child\")\n              .attr('class', 'rows')\n              .attr(\"transform\", \"translate(0,\"+contentTop+\")\");\n\n    if (labelsWidth > 0) {\n      labels = svg.append(\"g\")\n                  .attr('class', 'labels')\n                  .attr(\"transform\", \"translate(0,\"+contentTop+\")\");\n    }\n\n    content = svg.append(\"g\")\n               .attr('class', 'content')\n               .attr(\"transform\", \"translate(\" + labelsWidth + \",\"+contentTop+\")\");\n\n    content.append(\"g\")\n      .attr(\"class\", \"bars\");\n\n    if (scrubbable) {\n      scrubber = content.append(\"g\")\n                      .attr('class', 'scrubber');\n\n      scrubber.append('line')\n        .attr('x1', 0)\n        .attr('x2', 0)\n        .attr('y1', 0)\n        .attr('y2', contentHeight);\n\n      scrubber.append('text')\n        .attr('y', contentHeight);\n    }\n\n    if (scrubbable || selectable) {\n      svg.on('mousemove', function(){\n        var contentWidth = self.get('contentWidth'),\n            contentHeight = self.get('contentHeight');\n\n        var mouse = d3.mouse(content.node());\n        if (mouse[1] >= 0 && mouse[1] <= contentHeight) {\n          if (selectable) {\n            self.doHighlight(mouse[1]);\n          }\n\n          if (scrubbable) {\n            if (mouse[0] >= 0 && mouse[0] <= contentWidth) {\n              self.doScrub(mouse[0]);\n            }\n          }\n        }\n      });\n    }\n\n    if (selectable) {\n      svg.on('click', function() {\n        var rowItems = rows.selectAll('.row'),\n            y = d3.mouse(svg.node())[1];\n\n        rowItems.classed('selected', function(d, i) {\n          var top = Number(this.getAttribute('y')),\n              bottom = top + Number(this.getAttribute('height')),\n              isSelected = top <= y && y < bottom;\n\n          self._labelAt(i).classed('selected', isSelected);\n\n          return isSelected;\n        });\n\n        self.set('selection', rowItems.filter('.selected').data().map(function (item) {\n          return item.content || item;\n        }));\n      });\n    }\n\n    if (brushable) {\n      content.append(\"g\")\n        .attr(\"class\", \"x brush\");\n    }\n\n    this.renderNodes();\n\n    Ember.$(window).on('resize', Ember.$.proxy(this, 'windowDidResize'));\n    this.windowDidResize();\n  },\n\n  willDestroyElement: function() {\n    Ember.$(window).off('resize', Ember.$.proxy(this, 'windowDidResize'));\n  },\n\n  windowDidResize: function() {\n    this.notifyPropertyChange('maximumWidth');\n  },\n\n  init: function() {\n    this._super();\n    this.adjustXScaleRange();\n    this.adjustYScaleRange();\n    this.updateXAxisScale();\n  }\n});\n\n})();\n//@ sourceURL=ember-timetree/views/timetree_view");